# Generated by Django 3.1.1 on 2020-11-28 05:10

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Trace',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('action', models.PositiveSmallIntegerField(choices=[(1, 'Crear'), (2, 'Editar'), (3, 'Eliminar')], verbose_name='acción')),
                ('message', models.TextField(verbose_name='mensaje')),
                ('object_id', models.PositiveIntegerField(verbose_name='id del objecto')),
                ('date', models.DateTimeField(auto_now_add=True, verbose_name='fecha')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='contenido')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL, verbose_name='usuario')),
            ],
            options={
                'verbose_name': 'Rastro',
            },
        ),
        migrations.CreateModel(
            name='Rule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_date', models.DateTimeField(auto_now_add=True, verbose_name='fecha de creación')),
                ('created_user', models.CharField(editable=False, max_length=128, null=True, verbose_name='creado por')),
                ('modified_date', models.DateTimeField(auto_now=True, verbose_name='última fecha de modificación')),
                ('modified_user', models.CharField(editable=False, max_length=128, null=True, verbose_name='modificado por')),
                ('is_active', models.BooleanField(default=True, verbose_name='activo')),
                ('check_create', models.BooleanField(verbose_name='rastrear creaciones')),
                ('check_edit', models.BooleanField(verbose_name='rastrear modificaciones')),
                ('check_delete', models.BooleanField(verbose_name='rastrear eliminaciones')),
                ('content_type', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='Objeto')),
            ],
            options={
                'verbose_name': 'Regla',
            },
        ),
    ]



""" Audit model admin """

# Django
from django.contrib import admin

# Models
from .models import Trace, Rule

@admin.register(Trace)
class TraceAdmin(admin.ModelAdmin):
    list_display = ('content_object', 'date', 'user', 'action')
    list_filter = ("action", "date")

    def get_date(self, obj):
        return obj.date

@admin.register(Rule)
class RuleAdmin(admin.ModelAdmin):
    fields = ('content_type', 'check_create', 'check_edit', 'check_delete', 'is_active')
    list_display = ('content_type', 'check_create', 'check_edit', 'check_delete', 'created_user', 'modified_user')



""" Audit models. """

# Django
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.auth import get_user_model
from django.conf import settings
from django.db import models

# Models
User = get_user_model()


class BaseModel(models.Model):
    """Audit base model which every models inherit."""

    created_date = models.DateTimeField(
        auto_now_add=True, verbose_name="fecha de creación"
    )
    created_user = models.CharField(
        max_length=128, editable=False, null=True, verbose_name="creado por"
    )
    modified_date = models.DateTimeField(
        auto_now=True, verbose_name="última fecha de modificación"
    )
    modified_user = models.CharField(
        max_length=128, editable=False, null=True, verbose_name="modificado por"
    )
    is_active = models.BooleanField(default=True, verbose_name="activo")

    def _get_user(self, username):
        """Get user by provide username"""
        try:
            user = User.objects.get(username=username)
            return user
        except User.DoesNotExist:
            return username

    @property
    def creation_user(self):
        """Get created user"""
        return self._get_user(self.created_user)

    @property
    def modification_user(self):
        """Get updated user"""
        return self._get_user(self.modified_user)

    class Meta:
        """Meta options."""

        abstract = True


class Trace(models.Model):
    """ Model that stores audit logs according to the rules. """

    class ActionChoices(models.IntegerChoices):
        CREATE = 1, 'Crear'
        EDIT = 2, 'Editar'
        DELETE = 3, 'Eliminar'

    id = models.BigAutoField(primary_key=True)
    action = models.PositiveSmallIntegerField(
        choices=ActionChoices.choices, verbose_name="acción"
    )
    message = models.TextField(verbose_name='mensaje')
    content_type = models.ForeignKey(
        "contenttypes.ContentType", 
        on_delete=models.CASCADE,
        verbose_name='contenido'
    )
    object_id = models.PositiveIntegerField(
        verbose_name='id del objecto'
    )
    content_object = GenericForeignKey("content_type", "object_id")
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name='usuario'
    )
    date = models.DateTimeField(auto_now_add=True, verbose_name='fecha')

    def __str__(self):
        return f"{self.content_object}, {self.get_action_display()}"

    class Meta:
        verbose_name = "Rastro"


class Rule(BaseModel):
    """ Modelo que almacena las reglas de auditoría. """

    content_type = models.OneToOneField(
        "contenttypes.ContentType", on_delete=models.CASCADE, verbose_name="Objeto"
    )
    check_create = models.BooleanField("rastrear creaciones")
    check_edit = models.BooleanField("rastrear modificaciones")
    check_delete = models.BooleanField("rastrear eliminaciones")

    def __str__(self):
        return f"Regla de {str(self.content_type).lower()}"

    class Meta:
        verbose_name = "Regla"



""" Audit signals """

# Python
import json
from datetime import datetime

# Django
from django.forms.models import model_to_dict
from django.db.models.signals import post_save, post_delete, pre_save
from django.dispatch import receiver

# Middleware
from .middleware import TracingMiddleware

# Models
from .models import Trace, Rule, BaseModel


""" Util function """


def get_diff(instance, created):
    def prepare_dict(dict):
        for key in dict:
            dict[key] = tuple(dict[key]) if isinstance(dict[key], list) else dict[key]
        return dict

    presave = prepare_dict(TracingMiddleware.get_data())
    postsave = prepare_dict(model_to_dict(instance))

    diff = {key: value for key, value in presave.items() - postsave.items()}

    diff = postsave if created else diff
    return diff


"""Signals for reload rules"""


@receiver(post_save, sender=Rule)
def save_rule(sender, **kwargs):
    TracingMiddleware.reload_rules()


@receiver(post_delete, sender=Rule)
def delete_rule(sender, **kwargs):
    TracingMiddleware.reload_rules()


""" Signal for presave instance """


@receiver(pre_save)
def presave_log(sender, instance, **kwargs):
    try:
        presave = sender.objects.get(pk=instance.pk)
        data = model_to_dict(presave)
    except sender.DoesNotExist:
        data = {}
    TracingMiddleware.set_data(data)


""" Signal for postsave instance """


@receiver(post_save)
def save_log(sender, instance, created, **kwargs):
    rule = TracingMiddleware.get_rule_by_classname(sender._meta.model_name)
    if not rule:
        return

    diff = get_diff(instance, created)
    if not diff:
        return

    if created and rule.get("check_create"):
        message = {"creado": diff}
        options = {"action": Trace.ActionChoices.CREATE}
    elif not created and rule.get("check_edit"):
        message = {"editado": diff}
        options = {"action": Trace.ActionChoices.EDIT}
    else:
        return

    options.update(
        {
            "content_object": instance,
            "user": TracingMiddleware.get_user(),
            "message": json.dumps(message),
        }
    )
    Trace.objects.create(**options)


""" Signal for post_delete instance """


@receiver(post_delete)
def save_delete(sender, instance, **kwargs):
    rule = TracingMiddleware.get_rule_by_classname(sender._meta.model_name)

    if not rule:
        return

    if rule.get("check_delete"):
        message = {"eliminado": {}}
        options = {
            "action": Trace.ActionChoices.CREATE,
            "content_object": instance,
            "user": TracingMiddleware.get_user(),
            "message": json.dumps(message),
        }
        Trace.objects.create(**options)


""" """


@receiver(post_save)
def save_user_in_base_model(sender, instance, created, **kwargs):
    """Save audit fields"""
    if not issubclass(sender, BaseModel):
        return

    # Disconect signals
    pre_save.disconnect(presave_log)
    post_save.disconnect(save_log)
    post_save.disconnect(save_user_in_base_model)

    # Update user
    user = TracingMiddleware.get_user()
    user = user.username if user else "root"
    if created:
        instance.created_user = user
    instance.modified_user = user
    instance.save()

    # Reconnect sigals
    post_save.connect(save_user_in_base_model)
    post_save.connect(save_log)
    pre_save.connect(presave_log)



""" Tracing Middleware """

# Python
import threading

# Models
from .models import Rule


def load_rules():
    try:
        object_list = Rule.objects.filter(is_active=True).values(
            "content_type__model", "check_create", "check_edit", "check_delete"
        )
        rules = {
            obj.get("content_type__model"): {
                "check_create": obj.get("check_create"),
                "check_edit": obj.get("check_edit"),
                "check_delete": obj.get("check_delete"),
            }
            for obj in object_list
        }
    except:
        rules = {}
    print(f"Se cargaron {len(rules)} reglas.")
    return rules


class TracingMiddleware:
    thread_local = threading.local()
    rules = load_rules()

    def __init__(self, get_response):
        self.get_response = get_response

    @classmethod
    def set_data(cls, data):
        cls.thread_local.data = data

    @classmethod
    def get_data(cls):
        if hasattr(cls.thread_local, "data"):
            return cls.thread_local.data

    @classmethod
    def set_user(cls, user):
        cls.thread_local.user = user

    @classmethod
    def get_user(cls):
        if hasattr(cls.thread_local, "user"):
            return cls.thread_local.user

    @classmethod
    def reload_rules(cls):
        cls.rules = load_rules()

    @classmethod
    def get_rule_by_classname(cls, classname):
        if classname.lower() in cls.rules:
            return cls.rules.get(classname.lower())

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called
        self.thread_local.user = request.user

        if request.method == "POST":
            self.thread_local.data = request.POST.dict()
        response = self.get_response(request)
        
        return response
